from dbsamizdat import SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger


fruittable_SQL = """
    CREATE TABLE "Fruit" (
        id integer PRIMARY KEY,
        name varchar(100)
    );

    INSERT INTO "Fruit"
    SELECT
        *
    FROM (
        VALUES
            (1, 'banana'),
            (2, 'pear'),
            (3, 'apple'),
            (4, 'rambutan')
    ) AS thefruits;

    CREATE TABLE "Pet" (
        id integer PRIMARY KEY,
        name varchar(100)
    );

    INSERT INTO "Pet"
    SELECT
        *
    FROM (
        VALUES
            (1, 'ocelot'),
            (2, 'khoi carp'),
            (3, 'rolypoly'),
            (4, 'drosophila')
    ) AS thepets;
"""


class DealFruitView(SamizdatView):
    deps_on_unmanaged = {('public', 'Fruit')}
    sql_template = """
        ${preamble}
        SELECT name FROM "public"."Fruit" ORDER BY random() LIMIT 1;
        ${postamble}
    """


class DealFruitFun(SamizdatFunction):
    deps_on = {DealFruitView}
    sql_template = f"""
        ${{preamble}}
        RETURNS text AS
        $BODY$
        SELECT format('Have a piece of %s !', name) FROM {DealFruitView.db_object_identity}
        $BODY$
        LANGUAGE SQL
        IMMUTABLE
    """


class DealFruitFunWithName(SamizdatFunction):
    deps_on = {DealFruitFun}
    function_name = 'DealFruitFun'
    function_arguments_signature = 'name text'
    sql_template = f"""
        ${{preamble}}
        RETURNS text AS
        $BODY$
        SELECT format('Hey %s! %s', name, "{DealFruitFun.name}"())
        $BODY$
        LANGUAGE SQL
        IMMUTABLE
    """


class Treat(SamizdatMaterializedView):
    deps_on_unmanaged = {('public', 'Pet')}
    deps_on = {DealFruitFunWithName}
    refresh_triggers = {('public', 'Pet'), ('public', 'Fruit')}
    sql_template = f"""
        ${{preamble}}
        SELECT "{DealFruitFunWithName.name}"(name) AS treat FROM "public"."Pet"
        ${{postamble}};
    """


class PetUppercase(SamizdatMaterializedView):
    refresh_concurrently = True
    deps_on_unmanaged = {('public', 'Pet')}
    sql_template = """
        ${preamble}
        SELECT id, upper(name) FROM "public"."Pet"
        ${postamble};
        CREATE UNIQUE INDEX ON ${samizdatname} (id);
    """


class Raise(SamizdatFunction):
    sql_template = """
        ${preamble}
        RETURNS trigger AS
        $BODY$
        DECLARE arg_message text;
        BEGIN
            arg_message := TG_ARGV[0];
            RAISE EXCEPTION '%', arg_message;
        END;
        $BODY$
        LANGUAGE plpgsql IMMUTABLE;
    """


class FruitCensor(SamizdatTrigger):
    deps_on = {Raise}
    on_table = "Fruit"
    condition = "BEFORE INSERT OR UPDATE OF name"
    sql_template = """
        ${preamble}
        FOR EACH ROW
        WHEN (NEW.name ilike '%durian%')
        EXECUTE PROCEDURE "Raise"('Too delicious.')
    """



from collections import OrderedDict
from time import time as now
from string import Template
from abc import ABC, ABCMeta
from enum import Enum
from json import dumps as jsondumps
from hashlib import md5

from .util import db_object_identity, fqify_node, nodenamefmt
from .exceptions import UnsuitableNameError

_DBINFO_VERSION = 1  # Version number for signature format. For future use
PG_IDENTIFIER_MAXLEN = 63
PG_IDENTIFIER_VERBOTEN = set('"')  # Actually allowed, but we'd have to escape it everywhere. See https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
# Autogenerated triggers are numbered so as to make them run in samizdat dependency order when sorted alphabetically
# (as that's how PostgreSQL determines trigger run order). For this we need to leftpad the numbers (because 166 < 23, but 166 > 023, alphabetically)
# Sadly, this poses an artificial limit on the trigger "depth" and an exception will be raised if this is exceeded.
# The size here is chosen to retain short-ish trigger names without running into problems all too soon.
TRIGGER_DEPCOUNTER_PADDED_WIDTH = 5


class entitypes(Enum):
    VIEW = 'VIEW'
    MATVIEW = 'MATERIALIZED VIEW'
    FUNCTION = 'FUNCTION'
    TRIGGER = 'TRIGGER'


class SamizdatMeta(ABCMeta):
    @property
    def name(cls):
        return cls.__name__

    @property
    def fq(cls):
        return (cls.schema, cls.name)

    @property
    def db_object_identity(cls):
        return db_object_identity(cls.fq)

    def __str__(self):
        funcargs = getattr(self, 'function_arguments_signature', None)
        node = (self.schema, self.name, funcargs) if funcargs else (self.schema, self.name)
        return nodenamefmt(node)


class SamizdatFunctionMeta(SamizdatMeta, ABCMeta):
    """
    Since the identity of a functions is its name + some of its arguments (no OUT), in some form (sans defaults, types normalized,...), using the class name as the function name would be inconvenient
    due to class name clashes when defining multiple functions with same name but different arguments.
    Thus we use a separately defined `function_name` attribute.
    """

    @property
    def name(cls):
        return getattr(cls, 'function_name', cls.__name__)

    @property
    def fq(cls):
        return (cls.schema, cls.name, cls.function_arguments_signature)

    @property
    def creation_identity(cls):
        return '"%s"."%s"(%s)' % (cls.schema, cls.name, cls.creation_function_arguments())

    def creation_function_arguments(cls):
        return getattr(cls, 'function_arguments', cls.function_arguments_signature)


class SamizdatTriggerMeta(SamizdatMeta, ABCMeta):
    """
    A trigger is not directly associated with a schema (indirectly it is, through a table.)
    Yet trigger names do not need to be unique in a schema.
    Yet we need a unique identifier for them, within the dbsamizdat collection.
    So we use the table the trigger is attached to as a component in the fully qualified name.
    """
    @property
    def fq(cls):
        return ('.'.join(fqify_node(cls.on_table)), cls.name)

    def __str__(self):
        return nodenamefmt(self.fq)


class Samizdat(ABC):
    """
    Abstract parent class for dbsamizdat classes.
    """

    def __init__(self, *args, **kwargs):
        raise RuntimeError("""Error: Don't instantiate objects of type "%s".\nUse it as a static class.""" % self.__class__.__name__)

    deps_on = set()
    deps_on_unmanaged = set()
    schema = "public"
    sql_template = """
        -- There should be a class-dependent body for ${samizdatname} here.
        -- See README.md.
        """

    @classmethod
    def validate_name(cls):
        """
        Check whether name is a valid PostgreSQL identifier. Note that we'll quote the identifier everywhere ("quoted identifier" per https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS)
        """
        if any(map(lambda c: ord(c) > 0x7F, cls.name)):
            raise UnsuitableNameError('Name contains non-ASCII characters', samizdat=cls)  # Technically we could UESCAPE these, and make the length calculation much more complicated.
        if len(cls.name) > PG_IDENTIFIER_MAXLEN:
            raise UnsuitableNameError('Name is too long', samizdat=cls)
        if set(cls.name) & PG_IDENTIFIER_VERBOTEN:
            raise UnsuitableNameError(F'''Name contains unwelcome characters ({set(cls.name) & PG_IDENTIFIER_VERBOTEN})''', samizdat=cls)

    @classmethod
    def definition_hash(cls):
        try:
            return cls.implanted_hash  # if this thing has been read from the DB
        except AttributeError:
            use_attrs = ('sql_template', 'db_object_identity', 'creation_identity')
            return md5('|'.join(map(lambda attrname: getattr(cls, attrname, ''), use_attrs)).encode('utf-8')).hexdigest()

    @classmethod
    def fqify(cls, ref):
        if isinstance(ref, str):
            return (cls.schema, ref)
        if isinstance(ref, SamizdatMeta):
            return ref.fq
        if isinstance(ref, tuple):
            return ref
        raise ValueError("Not a string, tuple, or Samizdat: '%s'" % ref)

    @classmethod
    def fqdeps_on(cls):
        return set(map(cls.fqify, cls.deps_on))

    @classmethod
    def fqdeps_on_unmanaged(cls):
        return set(map(cls.fqify, cls.deps_on_unmanaged))

    @classmethod
    def dbinfo(cls):
        """
        Returns descriptor (json) for this object, to be stored in the database
        """
        samizdat_dbinfo = OrderedDict()
        samizdat_dbinfo_contents = OrderedDict()
        samizdat_dbinfo_contents['version'] = _DBINFO_VERSION
        samizdat_dbinfo_contents['created'] = int(now())
        samizdat_dbinfo_contents['definition_hash'] = cls.definition_hash()
        samizdat_dbinfo["dbsamizdat"] = samizdat_dbinfo_contents
        return jsondumps(samizdat_dbinfo)

    @classmethod
    def sign(cls, cursor):
        '''
        Generate COMMENT ON sql storing a signature
        We need the cursor to let psycopg2 properly escape our json-as-text-string.
        '''
        return cursor.mogrify("""COMMENT ON {cls.entity_type.value} {cls.db_object_identity} IS %s;""".format(cls=cls), (cls.dbinfo(),)).decode()

    @classmethod
    def create(cls):
        """
        SQL to create this DB object
        """
        subst = dict(
            preamble="""CREATE {cls.entity_type.value} {cls.db_object_identity} AS""".format(cls=cls),
            postamble={'MATVIEW': 'WITH NO DATA'}.get(cls.entity_type.name, ''),
            samizdatname=cls.db_object_identity
        )
        return Template(cls.sql_template).safe_substitute(subst)

    @classmethod
    def drop(cls, if_exists=False):
        """
        SQL to drop object. Cascade because we have no idea about the dependency tree.
        """
        return 'DROP {cls.entity_type.value} {if_exists} {cls.db_object_identity} CASCADE;'.format(cls=cls, if_exists={True: 'IF EXISTS'}.get(if_exists, ''))

    @classmethod
    def and_sidekicks(cls, counter=None):
        """
        On some classes, this will yield autogenerated sidekick classes
        """
        yield cls
        yield from ()

    @classmethod
    def head_id(cls):
        return hash((cls.schema, cls.name, cls.entity_type.name, cls.definition_hash()))


class SamizdatView(Samizdat, ABC, metaclass=SamizdatMeta):
    entity_type = entitypes.VIEW


class SamizdatFunction(Samizdat, ABC, metaclass=SamizdatFunctionMeta):
    entity_type = entitypes.FUNCTION
    function_arguments_signature = ''
    autorefresher = False  # this field serves to identify this node as autogenerated through a materialized view with populated `refresh_triggers` attribute

    @classmethod
    def create(cls):
        subst = dict(
            preamble="""CREATE {cls.entity_type.value} {cls.creation_identity}""".format(cls=cls),
            samizdatname=cls.db_object_identity
        )
        return Template(cls.sql_template).safe_substitute(subst)

    @classmethod
    def head_id(cls):
        return hash((cls.schema, cls.name, cls.entity_type.name, cls.definition_hash()))


class SamizdatTrigger(Samizdat, ABC, metaclass=SamizdatTriggerMeta):
    schema = None  # Triggers are only indirectly in a schema (through the table they're defined on.)
    entity_type = entitypes.TRIGGER
    on_table = None
    condition = None
    autorefresher = False  # this field serves to identify this node as autogenerated through a materialized view with populated `refresh_triggers` attribute

    @classmethod
    def fqdeps_on_unmanaged(cls):
        return {fqify_node(n) for n in cls.deps_on_unmanaged | {cls.on_table}}

    @classmethod
    def create(cls):
        subst = dict(
            preamble="""CREATE {cls.entity_type.value} "{cls.name}" {cls.condition} ON {target_table}""".format(cls=cls, target_table=db_object_identity(cls.on_table)),
            samizdatname=cls.name
        )
        return Template(cls.sql_template).safe_substitute(subst)

    @classmethod
    def drop(cls, if_exists=False):
        return 'DROP {cls.entity_type.value} {if_exists} {cls.name} ON {target_table} CASCADE;'.format(cls=cls, if_exists={True: 'IF EXISTS'}.get(if_exists, ''), target_table=db_object_identity(cls.on_table))

    @classmethod
    def sign(cls, cursor):
        return cursor.mogrify("""COMMENT ON {cls.entity_type.value} "{cls.name}" ON {on_table} IS %s;""".format(cls=cls, on_table=db_object_identity(cls.on_table)), (cls.dbinfo(),)).decode()

    @classmethod
    def head_id(cls):
        return hash((fqify_node(cls.on_table)[0], cls.name, cls.entity_type.name, fqify_node(cls.on_table)[1], cls.definition_hash()))


class SamizdatMaterializedView(Samizdat, ABC, metaclass=SamizdatMeta):
    entity_type = entitypes.MATVIEW
    refresh_concurrently = False
    refresh_triggers = set()
    AUTOREFRESHER_COUNTER = 'autorefresher'

    @classmethod
    def fqrefresh_triggers(cls):
        return set(map(cls.fqify, cls.refresh_triggers))

    @classmethod
    def refresh(cls, concurrent_allowed=True):
        return """REFRESH MATERIALIZED VIEW {concurrently} {cls.db_object_identity};""".format(cls=cls, concurrently={True: 'CONCURRENTLY'}.get(concurrent_allowed and cls.refresh_concurrently, ''))

    @classmethod
    def gen_refresh_triggerfunction(cls):
        """
        yields a SamizdatFunction; a trigger-returning function that refreshes this materialized view.
        """
        if cls.refresh_triggers:
            yield type(
                F'{cls.name}_refresh',
                (SamizdatFunction,),
                dict(
                    schema=cls.schema,  # The trigger function will be created in the same schema as the view. That's not a requirement, but we have to choose *something*.
                    deps_on={cls.fq},
                    autorefresher=True,
                    sql_template="""
                        ${preamble}
                        RETURNS trigger AS $THEBODY$
                        BEGIN
                        %s
                        RETURN NULL;
                        END;
                        $THEBODY$ LANGUAGE plpgsql;
                    """ % cls.refresh()
                )
            )

    @classmethod
    def gen_refresh_tabletriggers(cls, counter=None):
        """
        yields SamizdatTriggers on the tables, triggering a refresh this materialized view
        """
        dep_order = counter[cls.AUTOREFRESHER_COUNTER]
        if dep_order > 10**TRIGGER_DEPCOUNTER_PADDED_WIDTH - 1:
            raise ValueError(F"Trigger creation order {dep_order} requires a autonumbering padding width increase (TRIGGER_DEPCOUNTER_PADDED_WIDTH).")
        triggerfn = next(cls.gen_refresh_triggerfunction(), None)
        if triggerfn:
            for ix, triggertable in enumerate(sorted(cls.fqrefresh_triggers())):
                class_name = F't%.{TRIGGER_DEPCOUNTER_PADDED_WIDTH}d_%d_autorefresh' % (dep_order, ix)
                yield type(
                    class_name,
                    (SamizdatTrigger,),
                    dict(
                        on_table=triggertable,
                        condition="AFTER UPDATE OR INSERT OR DELETE OR TRUNCATE",
                        deps_on={triggerfn.fq},
                        autorefresher=True,
                        sql_template="""
                                ${preamble}
                                FOR EACH STATEMENT EXECUTE PROCEDURE %s;
                            """ % db_object_identity(triggerfn.fq)
                    )
                )

    @classmethod
    def and_sidekicks(cls, counter=None):
        yield cls
        if cls.refresh_triggers:
            counter.update({cls.AUTOREFRESHER_COUNTER: 1})
            yield from cls.gen_refresh_triggerfunction()
            yield from cls.gen_refresh_tabletriggers(counter=counter)



"""
API for using dbsamizdat as a library in Django
"""

from argparse import Namespace as _Namespace
from typing import Iterable, Union

from .runner import cmd_refresh as _cmd_refresh, cmd_sync as _cmd_sync, cmd_nuke as _cmd_nuke, txstyle
from .samizdat import Samizdat

_CMD_ARG_DEFAULTS = dict(
    in_django=True,
    verbosity=1,
    log_rather_than_print=True,
    samizdatmodules=tuple(),
)


def refresh(dbconn: str = 'default', transaction_style: txstyle = txstyle.JUMBO, belownodes: Iterable[Union[str, tuple, Samizdat]] = tuple()):
    """Refresh materialized views, in dependency order, optionally restricted to views depending directly or transitively on any of the DB objects specified in `belownodes`."""
    args = _Namespace(
        **_CMD_ARG_DEFAULTS,
        dbconn=dbconn,
        txdiscipline=transaction_style.value,
        belownodes=belownodes,
    )
    _cmd_refresh(args)


def sync(dbconn: str = 'default', transaction_style: txstyle = txstyle.JUMBO):
    """Sync dbsamizdat state to the DB."""
    args = _Namespace(
        **_CMD_ARG_DEFAULTS,
        dbconn=dbconn,
        txdiscipline=transaction_style.value,
    )
    _cmd_sync(args)


def nuke(dbconn: str = 'default', transaction_style: txstyle = txstyle.JUMBO):
    """Remove any database object tagged as samizdat."""
    args = _Namespace(
        **_CMD_ARG_DEFAULTS,
        dbconn=dbconn,
        txdiscipline=transaction_style.value,
    )
    _cmd_nuke(args)



def fqify_node(node):
    """
    normalize node names to (schema, nodename) format, assuming the 'public' schema for not-fully-qualified node names
    """
    if isinstance(node, str):
        firstpart, *rest = node.split('.', maxsplit=1)
        if rest:
            return (firstpart, rest[0])
        return ("public", firstpart)
    return node


def nodenamefmt(node):
    """
    format node for presentation purposes. If it's in the public schema, omit the "public" for brevity.
    """
    if isinstance(node, str):
        return node
    if isinstance(node, tuple):
        schema, name, *args = node
        identifier = f"{schema}.{name}" if schema not in {'public', None} else name
        if args and args[0]:
            return f'{identifier}({args[0]})'
        return identifier
    return str(node)  # then it should be a Samizdat


def db_object_identity(thing):
    schema, name, *fnargs = fqify_node(thing)
    args = f'({fnargs[0]})' if fnargs else ''
    return '"%s"."%s"%s' % (schema, name, args)


def sqlfmt(sql: str):
    return '\n'.join(('\t\t' + line for line in sql.splitlines()))



from typing import List

from .util import sqlfmt


class SamizdatException(Exception):
    def __init__(self, message, samizdat=None):
        self.message = message
        self.samizdat = samizdat

    def __str__(self):
        sd_subject = f'{repr(self.samizdat)} : ' if self.samizdat else ''
        return f'{sd_subject}{self.message}'


class NameClashError(SamizdatException):
    pass


class UnsuitableNameError(SamizdatException):
    pass


class DanglingReferenceError(SamizdatException):
    pass


class TypeConfusionError(SamizdatException):
    pass


class DatabaseError(SamizdatException):
    def __init__(self, message, dberror, samizdat, sql):
        self.message = message
        self.dberror = dberror
        self.samizdat = samizdat
        self.sql = sql

    def __str__(self):
        return f'''
            While executing:
            {sqlfmt(self.sql)}

            a DB error was raised:
            {self.dberror}

            while we were processing the samizdat:
            {repr(self.samizdat)}

            furthermore:
            {self.message}
            '''


class DependencyCycleError(SamizdatException):
    def __init__(self, message, samizdats):
        self.message = message
        self.samizdats = samizdats

    def __str__(self):
        sd_subjects = ', '.join(self.samizdats)
        return f'{sd_subjects} : {self.message}'


class FunctionSignatureError(SamizdatException):
    def __init__(self, samizdat, candidate_arguments: List[str]):
        self.samizdat = samizdat
        self.candidate_arguments = candidate_arguments

    def __str__(self):
        sd_subject = repr(self.samizdat)
        candidate_args = '\n'.join(self.candidate_arguments)
        args_herald = f"the following candidates:\n{candidate_args}" if len(self.candidate_arguments) > 1 else f'"{candidate_args}"'
        return f'''
            After executing:
            {sqlfmt(self.samizdat.create())}

            which we did in order to create the samizdat function:
            {sd_subject}

            we were not able to identify the resulting database function via its call signature of:
            {self.samizdat.db_object_identity}

            because, we figure, that is not actually the effective call signature resulting from the function arguments, which are:
            "({self.samizdat.function_arguments})"

            We queried the database to find out what the effective call argument signature should be instead, and came up with:
            {args_herald}

            HINT: Amend the {sd_subject} .function_arguments_signature and/or .function_arguments attributes.
            For more information, consult the README.'''



import argparse
import sys
from time import monotonic
from logging import getLogger
from enum import Enum

import psycopg2

from .libdb import dbinfo_to_class, dbstate_equals_definedstate, get_dbstate, DBObjectType
from .libgraph import depsort_with_sidekicks, node_dump, sanity_check, subtree_depends
from .graphvizdot import dot
from .loader import get_samizdats
from . import entitypes
from .exceptions import SamizdatException, DatabaseError, FunctionSignatureError
from .util import fqify_node, nodenamefmt, sqlfmt


class txstyle(Enum):
    CHECKPOINT = 'checkpoint'
    JUMBO = 'jumbo'
    DRYRUN = 'dryrun'


logger = getLogger(__name__)
PRINTKWARGS = dict(file=sys.stderr, flush=True)


def timer():
    last = monotonic()
    while True:
        cur = monotonic()
        yield (cur - last)
        last = cur


def vprint(args: argparse.Namespace, *pargs, **pkwargs):
    if args.log_rather_than_print:
        logger.info(' '.join(map(str, pargs)))
    elif args.verbosity:
        print(*pargs, **{**PRINTKWARGS, **pkwargs})


def vvprint(args: argparse.Namespace, *pargs, **pkwargs):
    if args.log_rather_than_print:
        logger.debug(' '.join(map(str, pargs)))
    elif args.verbosity > 1:
        print(*pargs, **{**PRINTKWARGS, **pkwargs})


def get_cursor(args):
    cursor = None
    if args.in_django:
        from django.db import connections
        cursor = connections[args.dbconn].cursor().cursor
    else:
        cursor = psycopg2.connect(args.dburl).cursor()
    cursor.execute('BEGIN;')  # And so it begins…
    return cursor


def txi_finalize(cursor, args):
    do_what = {txstyle.JUMBO.value: 'COMMIT;', txstyle.DRYRUN.value: 'ROLLBACK;'}.get(args.txdiscipline)
    if do_what:
        cursor.execute(do_what)


def cmd_refresh(args):
    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules)))
    matviews = [sd for sd in samizdats if sd.entity_type == entitypes.MATVIEW]

    if args.belownodes:
        rootnodes = {fqify_node(rootnode) for rootnode in args.belownodes}
        allnodes = node_dump(samizdats)
        if rootnodes - allnodes:
            raise ValueError('''Unknown rootnodes:\n\t- %s''' % '\n\t- '.join([nodenamefmt(rootnode) for rootnode in rootnodes - allnodes]))
        subtree_bundle = subtree_depends(samizdats, rootnodes)
        matviews = [sd for sd in matviews if sd in subtree_bundle]

    max_namelen = max(len(str(ds)) for ds in matviews)

    def refreshes():
        for sd in matviews:
            yield 'refresh', sd, sd.refresh(concurrent_allowed=True)
    executor(refreshes(), args, cursor, max_namelen=max_namelen, timing=True)
    txi_finalize(cursor, args)


def cmd_sync(args):
    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules)))
    issame, excess_dbstate, excess_definedstate = dbstate_equals_definedstate(cursor, samizdats)
    if issame:
        vprint(args, "No differences, nothing to do.")
        return
    max_namelen = max(len(str(ds)) for ds in excess_dbstate | excess_definedstate)
    if excess_dbstate:
        def drops():
            for sd in excess_dbstate:
                yield 'drop', sd, sd.drop(if_exists=True)  # we don't know the deptree; so they may have vanished through a cascading drop of a previous object
        executor(drops(), args, cursor, max_namelen=max_namelen, timing=True)
    if excess_definedstate:
        def creates():
            to_create_ids = {sd.head_id() for sd in excess_definedstate}
            for sd in samizdats:  # iterate in proper creation order
                if sd.head_id() in to_create_ids:
                    yield 'create', sd, sd.create()
                    yield 'sign', sd, sd.sign(cursor)
        executor(creates(), args, cursor, max_namelen=max_namelen, timing=True)

        matviews_to_refresh = {sd.head_id() for sd in excess_definedstate if sd.entity_type == entitypes.MATVIEW}
        if matviews_to_refresh:
            def refreshes():
                for sd in samizdats:  # iterate in proper creation order
                    if sd.head_id() in matviews_to_refresh:
                        yield 'refresh', sd, sd.refresh(concurrent_allowed=False)
            executor(refreshes(), args, cursor, max_namelen=max_namelen, timing=True)
    txi_finalize(cursor, args)


def cmd_diff(args):
    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules)))
    issame, excess_dbstate, excess_definedstate = dbstate_equals_definedstate(cursor, samizdats)
    if issame:
        vprint(args, "No differences.")
        exit(0)

    max_namelen = max(len(str(ds)) for ds in excess_dbstate | excess_definedstate)

    def statefmt(state, prefix):
        return '\n'.join(f'%s%-17s\t%-{max_namelen}s\t%s' % (prefix, sd.entity_type.value, sd, sd.definition_hash()) for sd in sorted(state, key=lambda sd: str(sd)))
    if excess_dbstate:
        vprint(args, statefmt(excess_dbstate, 'Not in samizdats:\t'), file=sys.stdout)
    if excess_definedstate:
        vprint(args, statefmt(excess_definedstate, 'Not in database:   \t'), file=sys.stdout)
    exit(100 + (1 if excess_dbstate else 0 | 2 if excess_definedstate else 0))


def cmd_printdot(args):
    print('\n'.join(dot(depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules))))))


def cmd_nuke(args, samizdats=None):
    cursor = get_cursor(args)

    def nukes():
        nonlocal samizdats
        if samizdats is None:
            samizdats = map(dbinfo_to_class, filter(lambda a: a[-1] is not None, get_dbstate(cursor)))
        for sd in samizdats:
            yield ("nuke", sd, sd.drop(if_exists=True))

    executor(nukes(), args, cursor)
    txi_finalize(cursor, args)


def executor(yielder, args, cursor, max_namelen=0, timing=False):

    action_timer = timer()

    def progressprint(ix, action_totake, sd, sql):
        if args.verbosity:
            if ix:
                # print the processing time of the *previous* action
                vprint(args, '%.2fs' % next(action_timer) if timing else '')
            vprint(args, f'%-7s %-17s %-{max_namelen}s ...' % (action_totake, sd.entity_type.value, sd), end='')
            vvprint(args, f'\n\n{sqlfmt(sql)}\n\n')

    action_cnt = 0
    for ix, progress in enumerate(yielder):
        action_cnt += 1
        progressprint(ix, *progress)
        action_totake, sd, sql = progress
        try:
            try:
                cursor.execute("BEGIN;")  # harmless if already in a tx
                cursor.execute(f"SAVEPOINT action_{action_totake};")
                cursor.execute(sql)
            except psycopg2.errors.UndefinedFunction as ouch:
                if action_totake == 'sign':
                    cursor.execute(f"ROLLBACK TO SAVEPOINT action_{action_totake};")  # get back to a non-error state
                    candidate_args = [c[3] for c in get_dbstate(cursor, which=DBObjectType.FOREIGN, entity_types=(entitypes.FUNCTION,)) if c[:2] == (sd.schema, sd.function_name)]
                    raise FunctionSignatureError(sd, candidate_args)
                raise ouch
        except psycopg2.Error as dberr:
            raise DatabaseError(f"{action_totake} failed", dberr, sd, sql)
        cursor.execute(f'RELEASE SAVEPOINT action_{action_totake};')
        if args.txdiscipline == txstyle.CHECKPOINT.value and action_totake != 'create':
            # only commit *after* signing, otherwise if later the signing somehow fails we'll have created an orphan DB object that we don't recognize as ours
            cursor.execute("COMMIT;")

    if action_cnt:
        vprint(args, '%.2fs' % next(action_timer) if timing else '')


def augment_argument_parser(p, in_django=False, log_rather_than_print=True):

    def perhaps_add_modules_argument(parser):
        if not in_django:
            parser.add_argument('samizdatmodules', nargs='+', help="Names of modules containing Samizdat subclasses")

    def add_dbarg_argument(parser):
        if in_django:
            parser.add_argument('dbconn', nargs='?', default='default', help="Django DB connection key (default:'default'). If you don't know what this is, then you don't need it.")
        else:
            parser.add_argument('dburl', help="PostgreSQL DB connection string. Trivially, this might be 'postgresql:///mydbname'. See https://www.postgresql.org/docs/14/static/libpq-connect.html#id-1.7.3.8.3.6 .")

    def add_txdiscipline_argument(parser):
        parser.add_argument('--txdiscipline', '-t', choices=(txstyle.CHECKPOINT.value, txstyle.JUMBO.value, txstyle.DRYRUN.value), default=txstyle.CHECKPOINT.value, help=f"""Transaction discipline. The "{txstyle.CHECKPOINT.value}" level commits after every dbsamizdat-level action. The safe default of "{txstyle.JUMBO.value}" creates one large transaction. "{txstyle.DRYRUN.value}" also creates one large transaction, but rolls it back.""")

    p.set_defaults(
        **dict(
            func=lambda whatevs: p.print_help(),
            in_django=in_django,
            log_rather_than_print=log_rather_than_print,
            samizdatmodules=[],
            verbosity=1,
        )
    )
    if not in_django:
        p.add_argument('--quiet', '-q', help="Be quiet (minimal output)", action="store_const", const=0, dest='verbosity')
        p.add_argument('--verbose', '-v', help="Be verbose (on stderr).", action="store_const", const=2, dest='verbosity')
    else:
        p.add_argument('-v', '--verbosity', default=1, type=int)
    subparsers = p.add_subparsers(title='commands')

    p_nuke = subparsers.add_parser('nuke', help='Drop all dbsamizdat database objects.')
    p_nuke.set_defaults(func=cmd_nuke)
    add_txdiscipline_argument(p_nuke)
    add_dbarg_argument(p_nuke)

    p_printdot = subparsers.add_parser('printdot', help='Print DB object dependency tree in GraphViz format.')
    p_printdot.set_defaults(func=cmd_printdot)
    perhaps_add_modules_argument(p_printdot)

    p_diff = subparsers.add_parser('diff', help='Show differences between dbsamizdat state and database state. Exits nonzero if any are found: 101 when there are excess DB-side objects, 102 if there are excess python-side objects, 103 if both sides have excess objects.')
    p_diff.set_defaults(func=cmd_diff)
    add_dbarg_argument(p_diff)
    perhaps_add_modules_argument(p_diff)

    p_refresh = subparsers.add_parser('refresh', help='Refresh materialized views, in dependency order')
    p_refresh.set_defaults(func=cmd_refresh)
    add_txdiscipline_argument(p_refresh)
    add_dbarg_argument(p_refresh)
    perhaps_add_modules_argument(p_refresh)
    p_refresh.add_argument('--belownodes', '-b', nargs='*', help="Limit to views that depend on ENTITYNAMES (usually, specific tables)", metavar='ENTITYNAMES')

    p_sync = subparsers.add_parser('sync', help='Make it so!')
    p_sync.set_defaults(func=cmd_sync)
    add_txdiscipline_argument(p_sync)
    add_dbarg_argument(p_sync)
    perhaps_add_modules_argument(p_sync)


def main():
    p = argparse.ArgumentParser(description='dbsamizdat, the blissfully naive PostgreSQL database object manager.')
    augment_argument_parser(p, log_rather_than_print=False)
    args = p.parse_args()
    try:
        args.func(args)
    except SamizdatException as argh:
        exit(f'\n\n\nFATAL: {argh}')
    except KeyboardInterrupt:
        exit('\nInterrupted.')


if __name__ == '__main__':
    main()



from logging import getLogger
from importlib import import_module
import inspect
from itertools import chain

from .samizdat import SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger, SamizdatMeta, SamizdatFunctionMeta, SamizdatTriggerMeta


logger = getLogger(__name__)

AUTOLOAD_MODULENAME = "dbsamizdat_defs"


def get_samizdats(modulelist=None, only_that_modulelist=False):

    def issamizdat(thing):
        excluded_classes = {SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger}
        return inspect.isclass(thing) and isinstance(thing, (SamizdatMeta, SamizdatFunctionMeta, SamizdatTriggerMeta)) and (thing not in excluded_classes)

    sdmodules = [import_module(sdmod) for sdmod in (modulelist or [])]
    if not only_that_modulelist:
        try:
            # if we're running in Django, we can try to autoload things
            from django.core.exceptions import ImproperlyConfigured
            from django.conf import settings
            from django.apps import apps
        except ImportError:
            pass  # no Django
        else:
            try:
                django_sdmodules = [import_module(sdmod) for sdmod in getattr(settings, 'DBSAMIZDAT_MODULES', [])]
                for appconfig in apps.get_app_configs():
                    try:
                        django_sdmodules.append(import_module('{}.{}'.format(appconfig.module.__package__, AUTOLOAD_MODULENAME)))
                    except ImportError as err:
                        if not err.msg.endswith(f"{AUTOLOAD_MODULENAME}'"):
                            raise err
                if not django_sdmodules:
                    logger.warn("""No settings.DBSAMIZDAT_MODULES defined, and none of your apps contain any "dbsamizdat_defs" module to autoload.""")
                sdmodules += django_sdmodules
            except ImproperlyConfigured:
                # assume we're not running in a fully booted Django
                pass

    return set((c for cname, c in chain(*map(lambda m: inspect.getmembers(m, issamizdat), sdmodules))))



from argparse import Namespace

from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import pre_migrate, post_migrate
from django.db.migrations import RunSQL, RunPython, AddField, RemoveField, RenameField, AlterField, DeleteModel, RenameModel, AlterModelTable
from django.contrib.contenttypes.management import RenameContentType

from .libgraph import unmanaged_refs, depsort_with_sidekicks, sanity_check, subtree_depends
from .loader import get_samizdats
from .runner import cmd_sync, cmd_nuke, txstyle, get_cursor
from .libdb import dbstate_equals_definedstate
from .util import fqify_node

DBCONN = 'default'  # Only migrations on the default DB connections are supported. For now.
SMART_MIGRATIONS = getattr(settings, 'DBSAMIZDAT_SMART_MIGRATIONS', False)  # Don't use with custom Operations!


def get_cmd_args(**kwargs):
    return Namespace(
        printprogress=True,
        verbosity=kwargs['verbosity'],
        dbconn=DBCONN,
        txdiscipline=txstyle.JUMBO.value,
        in_django=True,
        samizdatmodules=tuple(),
        log_rather_than_print=not kwargs.get('interactive'),
    )


def sync(**kwargs):
    cmd_sync(get_cmd_args(**kwargs))


def nuke(samizdats=None, **kwargs):
    cmd_nuke(get_cmd_args(**kwargs), samizdats=samizdats)


def get_django_cursor():
    return get_cursor(
        Namespace(
            in_django=True,
            dbconn=DBCONN,
        )
    )


def tables_affected_by(apps, plan):
    """
    Returns tables potentially affected by a migration plan.
    There are false positives, because we don't know what a view depending on a table exactly reads from that table.
    Worse, there may be false negatives, as we may not understand all types of migrations, and the migration plan is not a public API.
    """
    tables_affected = set()
    for mig, reverse in plan:
        if reverse:
            return (True, tables_affected)  # Too hard to think about, not too common, just nuke & recreate
        for operation in mig.operations:
            if isinstance(operation, RenameContentType):
                break  # sidekick operation of RenameModel, which we've already handled
            if isinstance(operation, (RunSQL, RunPython)):
                # Those could affect anything. We'll have to remove all our state.
                # No sense in analyzing any other migrations then, either; return early.
                return (True, tables_affected)
            try:
                model_meta = apps.get_model(mig.app_label, getattr(operation, 'model_name', getattr(operation, 'old_name', getattr(operation, 'name'))))._meta
            except (AttributeError, LookupError):
                break
            if not model_meta.managed:
                break  # as then the migration will not actually do anything
            if isinstance(operation, (AddField, RemoveField, RenameField, AlterField, DeleteModel, RenameModel, AlterModelTable)):
                tables_affected.add(fqify_node(model_meta.db_table))
    return (False, tables_affected)


def premigrate_handler(sender, **kwargs):
    if not (kwargs.get('plan')) or (kwargs['using'] != DBCONN):
        return

    if not SMART_MIGRATIONS:
        # Make no effort to determine whether we can be selective in which state we'll remove.
        nuke(**kwargs)
        return

    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats()))
    is_synced, *_whatevs = dbstate_equals_definedstate(get_django_cursor(), samizdats)
    if not is_synced:
        # There's unsynced samizdat state, and we can't tell if
        # a) new state depends on the post-migrate state or the pre-migrate state
        # b) old state will get in the way of the migrations
        # (or both.) Therefor, we can't sync, but neither can we leave any stuff lying around.
        # Thus we'll have to nuke DB state.
        nuke(**kwargs)
        return

    needs_wipe, tables_affected = tables_affected_by(kwargs['apps'], kwargs['plan'])
    if needs_wipe:
        nuke(**kwargs)
        return

    deps_affected = tables_affected & unmanaged_refs(samizdats)
    sds_affected = subtree_depends(samizdats, deps_affected)
    nuke(samizdats=sds_affected, **kwargs)


def postmigrate_handler(sender, **kwargs):
    if kwargs.get('plan') and kwargs['using'] == DBCONN:
        # Stuff went down, we may have dropped some things, sync will figure it out
        if kwargs.get('interactive'):
            print("Syncing DBSamizdat:")
        sync(**kwargs)


class DBSamizdatConfig(AppConfig):
    name = 'dbsamizdat'

    def ready(self):
        pre_migrate.connect(premigrate_handler, sender=self)
        post_migrate.connect(postmigrate_handler, sender=self)



from .util import nodenamefmt
from .libgraph import gen_autorefresh_edges, gen_edges, gen_unmanaged_edges, unmanaged_refs, toposort
from . import entitypes


def dot(samizdats):
    """
    Adorned GraphViz dotfile output.
    """
    samizdats = set(filter(lambda s: not getattr(s, 'autorefresher', False), samizdats))  # filter out autogenerated functions/triggers — autorefresh relationships will be expressed as edges only
    unmanaged = unmanaged_refs(samizdats)
    samizdat_map = {sd.fq: sd for sd in samizdats}
    topsorted = list(toposort({k: v.fqdeps_on() for k, v in samizdat_map.items()}))
    samerank = slice(-1, 1)

    def nodefmt(sd):
        styles = {
            entitypes.MATVIEW.name: '"%s" [shape=box3d fillcolor=red]',
            entitypes.VIEW.name: '"%s" [shape=box fillcolor=grey]',
            entitypes.FUNCTION.name: '"%s" [shape=hexagon fillcolor=olivedrab1]',
            entitypes.TRIGGER.name: '"%s" [shape=cds fillcolor=darkorchid1]',
        }
        return styles[sd.entity_type.name] % sd

    def nice_edges(edges):
        for nodepair in edges:
            yield tuple(map(nodenamefmt, nodepair))

    def enquote(thing):
        return '"%s"' % thing

    unmanaged_fmted = {nodenamefmt(n) for n in unmanaged}
    yield '''
    digraph {
    {
    node [style="filled" fontname="sans-serif" margin="0.15, 0.15"];
    newrank="true";
    '''
    yield '\n'.join(('"%s" [shape=house fillcolor=yellow]' % n for n in unmanaged_fmted))
    yield '\n'.join(map(nodefmt, samizdats))
    yield '}'

    yield '\n'.join(('"%s" -> "%s"' % e for e in nice_edges(gen_edges(samizdats))))
    yield '\n'.join(('"%s" -> "%s" [arrowhead="dot" color="navy" style="dashed" label="🗘" fontname="sans-serif" fontcolor="navy"]' % e for e in nice_edges(gen_autorefresh_edges(samizdats))))
    yield '\n'.join(('"%s" -> "%s" [color="red"]' % e for e in nice_edges(gen_unmanaged_edges(samizdats))))

    if unmanaged:
        yield '{ rank=min; %s }' % ' '.join(map(enquote, unmanaged_fmted))
        samerank = slice(0, -1)
    for rank in topsorted[samerank]:
        yield '{ rank=same; %s }' % ' '.join(map(enquote, map(nodenamefmt, rank)))
    yield '{ rank=max; %s }' % ' '.join(map(enquote, map(nodenamefmt, topsorted[-1])))
    yield '}'



from collections import Counter
from functools import reduce
from itertools import chain
from operator import or_

from toposort import toposort, CircularDependencyError

from . import entitypes
from .exceptions import NameClashError, DanglingReferenceError, TypeConfusionError, DependencyCycleError


def gen_edges(samizdats):
    for sd in samizdats:
        for n2 in sd.fqdeps_on():
            yield (n2, sd.fq)


def gen_autorefresh_edges(samizdats):
    for sd in filter(lambda sd: sd.entity_type == entitypes.MATVIEW, samizdats):
        for n2 in sd.refresh_triggers:
            yield (n2, sd.fq)


def gen_unmanaged_edges(samizdats):
    for sd in samizdats:
        for n2 in sd.fqdeps_on_unmanaged():
            yield (n2, sd.fq)


def node_dump(samizdats):
    """
    All nodes (managed or unmanaged)
    """
    return reduce(or_, (sd.fqdeps_on_unmanaged() | {sd.fq} for sd in samizdats))


def unmanaged_refs(samizdats):
    """
    All unmanaged nodes referenced
    """
    return set(reduce(or_, (sd.fqdeps_on_unmanaged() | (sd.fqrefresh_triggers() if (sd.entity_type == entitypes.MATVIEW) else set()) for sd in samizdats)))


def subtree_nodes(samizdats, subtree_root):
    """
    All nodes depending on subtree_root (includes subtree_root)
    """
    def stn(subtree_root):
        yield subtree_root
        revdeps = (sd.fq for sd in samizdats if (subtree_root in (sd.fqdeps_on() | sd.fqdeps_on_unmanaged())))
        yield from chain.from_iterable(map(stn, revdeps))

    return set(stn(subtree_root))


def subtree_depends(samizdats, roots):
    """
    Samizdats directly or indirectly depending on any root in roots
    """
    sdmap = {sd.fq: sd for sd in samizdats}
    return reduce(or_, (set(filter(None, (sdmap.get(name) for name in subtree_nodes(samizdats, rootnode)))) for rootnode in roots), set())


def depsort(samizdats):
    """
    Topologically sort samizdats
    """
    samizdat_map = {sd.fq: sd for sd in samizdats}
    depmap = {sd.fq: sd.fqdeps_on() for sd in samizdats}
    return [samizdat_map[name] for name in chain(*(sorted(level) for level in toposort(depmap)))]


def depsort_with_sidekicks(samizdats):
    counter = Counter()
    return list(chain(*(sd.and_sidekicks(counter) for sd in depsort(samizdats))))


def sanity_check(samizdats):
    for sd in samizdats:
        sd.validate_name()
    sd_fqs = set(sd.fq for sd in samizdats)
    sd_deps = set(chain(*(sd.fqdeps_on() for sd in samizdats)))
    sd_deps_unmanaged = set(chain(*(sd.deps_on_unmanaged for sd in samizdats)))

    # are there any classes with ambiguous DB identity?
    if len(sd_fqs) < len(samizdats):
        cnt = Counter((sd.db_object_identity for sd in samizdats))
        nonunique = [db_id for db_id, count in cnt.items() if count > 1]
        if nonunique:
            raise NameClashError('Non-unique DB entities specified: %s' % nonunique)
    # check if all declared samizdat deps are present
    if not sd_deps.issubset(sd_fqs):
        raise DanglingReferenceError('Nonexistent dependencies referenced: %s' % (sd_deps - sd_fqs))
    # assert none of the declared unmanaged deps are declared samizdat
    confused = sd_deps_unmanaged.intersection(sd_fqs)
    if confused:
        raise TypeConfusionError('Samizdat entity is also declared as *unmanaged* dependency: %s' % confused)
    # cycle detection - top level
    selfreffaulty = {sd for sd in samizdats if sd.fq in sd.deps_on}
    if selfreffaulty:
        raise DependencyCycleError('Self-referential dependency', (selfreffaulty.pop(),))
    # cycle detection - other levels; toposort will raise an exception if there's one
    sdfqmap = {sd.fq: sd for sd in samizdats}
    try:
        _ = depsort_with_sidekicks(samizdats)
    except CircularDependencyError as ouch:
        cyclists = tuple(map(sdfqmap.get, ouch.data.keys()))
        raise DependencyCycleError('Dependency cycle detected', cyclists)
    return samizdats



from json import loads as jsonloads
from typing import Iterable
from itertools import chain
from enum import IntFlag

from . import entitypes, SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger

COMMENT_MAGIC = """{"dbsamizdat": {"version":"""


class DBObjectType(IntFlag):
    SAMIZDAT = 1
    FOREIGN = 2


def get_dbstate(cursor, which: DBObjectType = DBObjectType.SAMIZDAT, entity_types: Iterable[entitypes] = (entitypes.VIEW, entitypes.MATVIEW, entitypes.FUNCTION, entitypes.TRIGGER)):
    """
    Capture and annotate the current DB state (functions, views and triggers)
    """

    def execfetch(sql):
        cursor.execute(sql)
        return cursor.fetchall()

    pg11_or_lower = cursor.connection.server_version < 110000
    function_filter = "NOT (p.proisagg OR p.proiswindow OR p.prosecdef)" if pg11_or_lower else "p.prokind NOT IN ('a', 'w', 'p')"

    fetches = {
        entitypes.VIEW: """
            SELECT n.nspname AS schemaname,
                c.relname AS viewname,
                'VIEW' as viewtype,
                pg_catalog.obj_description(c.oid, 'pg_class') AS commentcontent
            FROM pg_catalog.pg_class c
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE c.relkind = 'v'
                AND n.nspname <> 'pg_catalog'
                AND n.nspname <> 'information_schema'
                AND n.nspname !~ '^pg_toast'
            """,
        entitypes.MATVIEW: """
            SELECT n.nspname AS schemaname,
                c.relname AS viewname,
                'MATVIEW' as viewtype,
                pg_catalog.obj_description(c.oid, 'pg_class') AS commentcontent
            FROM pg_catalog.pg_class c
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
            WHERE c.relkind = 'm'
                AND n.nspname <> 'pg_catalog'
                AND n.nspname <> 'information_schema'
                AND n.nspname !~ '^pg_toast'
            """,
        entitypes.FUNCTION: f"""
            SELECT n.nspname AS "schemaname",
                p.proname AS "functionname",
                'FUNCTION',
                pg_catalog.pg_get_function_identity_arguments(p.oid) AS args,
                pg_catalog.obj_description(p.oid, 'pg_proc') AS commentcontent
            FROM pg_catalog.pg_proc p
            LEFT JOIN pg_catalog.pg_namespace n ON n.oid = p.pronamespace
            WHERE {function_filter}
                AND n.nspname <> 'pg_catalog'
                AND n.nspname <> 'information_schema'
            """,
        entitypes.TRIGGER: """
            SELECT
                pn.nspname AS schemaname,
                pt.tgname AS triggername,
                'TRIGGER',
                pc.relname AS tablename,
                pg_catalog.obj_description(pt.oid, 'pg_trigger') AS commentcontent
            FROM
                pg_trigger pt
                LEFT JOIN pg_class pc ON pt.tgrelid = pc.oid
                LEFT JOIN pg_catalog.pg_namespace pn ON pn.oid = pc.relnamespace
            WHERE
                pt.tgisinternal = False
            """
    }

    for *stuff, jinfo in chain(*map(execfetch, map(fetches.get, entity_types))):
        objtype = DBObjectType.SAMIZDAT if (jinfo and jinfo.startswith(COMMENT_MAGIC)) else DBObjectType.FOREIGN
        if objtype & which:
            definition_hash = None
            if objtype == DBObjectType.SAMIZDAT:
                meta = jsonloads(jinfo)['dbsamizdat']
                hashattr = {0: 'sql_template_hash', 1: 'definition_hash'}[meta['version']]
                definition_hash = meta[hashattr]
            yield tuple(stuff + [definition_hash])


def dbinfo_to_class(dbstate_info):
    """
    Reconstruct a class out of information found in the DB
    """
    typemap = {c.entity_type: c for c in (SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger)}
    schema, objectname, objecttype, *maybe_args, definition_hash = dbstate_info
    entity_type = entitypes[objecttype]
    classfields = dict(
        schema=schema,
        implanted_hash=definition_hash,
    )
    if entity_type == entitypes.FUNCTION:
        classfields.update(dict(
            function_arguments_signature=maybe_args[0],
            function_name=objectname,
        ))
    elif entity_type == entitypes.TRIGGER:
        classfields.update(dict(
            schema=None,
            on_table=(schema, maybe_args[0]),
        ))

    return type(objectname, (typemap[entitypes[objecttype]],), classfields)


def dbstate_equals_definedstate(cursor, samizdats):
    dbstate = {ds.head_id(): ds for ds in map(dbinfo_to_class, get_dbstate(cursor))}
    definedstate = {ds.head_id(): ds for ds in samizdats}
    return (
        dbstate.keys() == definedstate.keys(),
        {dbstate[k] for k in dbstate.keys() - definedstate.keys()},
        {definedstate[k] for k in definedstate.keys() - dbstate.keys()}
    )
